/*
 * Copyright (C) 2014 Erik-Oliver Blass
 *                    Travis Mayberry
 *                    Guevara Noubir
 *                    Kaan Onarlioglu <onarliog@ccs.neu.edu>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * hive.c, ver.2014.11.03
 *
 * HiVE core.
 */

#include "utils.h"
#include "crypto.h"
#include "hive.h"

/*
 * Size computation helpers.
 */

/* No of total data sectors on HiVE device. */
u64 get_data_sectors(u64 sectors) {
    return sectors * 4;
}

/* No of metadata sectors on HiVE device. */
u64 get_meta_sectors(u64 sectors) {
    return div_ceil(4 * (16 + (sizeof(sectors) * 2)) * sectors, HIVE_SECTOR_LEN);
}

/* No of translation map sectors on HiVE device. */
u64 get_map_sectors(u64 sectors) {
    return div_ceil(sizeof(sectors) * sectors, HIVE_SECTOR_LEN);
}

/* No of stash sectors on HiVE device. */
u64 get_stash_sectors() {
    return 2 * (HIVE_STASH_LEN + HIVE_STASH_MAP_LEN);
}

/* Total size of the data HiVE. */
u64 get_data_hive_len(u64 sectors) {
    return HIVE_HEADER_LEN + get_data_sectors(sectors) + get_meta_sectors(sectors) + get_stash_sectors();
}

/* Total size of the map HiVE. */
u64 get_map_hive_len(u64 sectors) {
    return HIVE_HEADER_LEN + get_data_sectors(sectors) + get_meta_sectors(sectors) + get_stash_sectors() +
        2 * get_map_sectors(sectors);
}

/*
 * This computes and returns the number of sectors that exceed the total device
 * size after rounding up sector counts to teh closest integers.
 */
unsigned get_size_correction(u64 device_sectors, u64 sectors_data_hive, u64 sectors_map_hive) {

    u64 total_len;
    int correction;

    total_len = get_data_hive_len(sectors_data_hive) + get_map_hive_len(sectors_map_hive);

    correction = total_len - device_sectors;
    return (correction > 0) ? div_ceil(correction, 4) : 0;
}

/*
 * Key derivation and management.
 */

/* Prompts the user for passwords and generates/derives the crypto keys. */
void get_keys(int op, struct hive_header_entry *she) {

    char *pass;
    char *pass_v;
    char hive_pass_keys[HIVE_MAX_WORLDS][HIVE_KEY_LEN];
    unsigned i;
    struct termios old_term;
    struct termios new_term;

    /* Turn terminal echo off. */
    tcgetattr(STDIN_FILENO, &old_term);
    new_term = old_term;
    new_term.c_lflag &= ~(ECHO);
    tcsetattr(STDIN_FILENO, TCSANOW, &new_term);

    /* Prompt for passwords. */
    for (i = 0; i < hive_worlds; ++i) {
        printf("Please enter password for Volume %u: ", i);
        if (scanf("%ms", &pass) != 1) {
            tcsetattr(STDIN_FILENO, TCSANOW, &old_term);
            die("\nError reading password!\n");
        }

        /* Check again. */
        if (op == HIVE_CREATE) {
            printf("\nPlease re-enter the password for Volume %u: ", i);
            if (scanf("%ms", &pass_v) != 1) {
                tcsetattr(STDIN_FILENO, TCSANOW, &old_term);
                memset(pass, 0, strlen(pass));
                die("\nError reading password!\n");
            }

            if (strcmp(pass, pass_v) != 0) {
                tcsetattr(STDIN_FILENO, TCSANOW, &old_term);
                memset(pass, 0, strlen(pass));
                memset(pass_v, 0, strlen(pass_v));
                die("\nPasswords do not match!\n");
            }

            memset(pass_v, 0, strlen(pass_v));
            free(pass_v);
        }

        /* Derive the pass keys. */
        if (op == HIVE_CREATE) {
            get_random_data(she[i].pass_salt, HIVE_SALT_LEN);
        }
        generate_key(pass, HIVE_KEY_LEN, hive_pass_keys[i], she[i].pass_salt, HIVE_SALT_LEN);

        memset(pass, 0, strlen(pass));
        free(pass);

        printf("\n");
    }

    /* Restore terminal settings. */
    tcsetattr(STDIN_FILENO, TCSANOW, &old_term);

    if (op == HIVE_CREATE) {

        /* Randomly generate the actual encryption keys for each volume. */
        for (i = 0; i < HIVE_MAX_WORLDS; ++i) {
            get_random_data(hive_keys[i], HIVE_KEY_LEN);

            if (i < hive_worlds) {
                /* Store encrypted encryption keys in the headers. */
                encrypt(hive_keys[i], she[i].key, HIVE_KEY_LEN, hive_pass_keys[i], 0);
                get_random_data(she[i].key_salt, HIVE_SALT_LEN);
                digest_key(hive_keys[i], HIVE_KEY_LEN, she[i].key_digest, HIVE_DIGEST_LEN, she[i].key_salt, HIVE_SALT_LEN);
            }
            else {
                /*
                 * If this is a single-world HiVE, fill in random values for the
                 * non-existent second world.
                 */
                get_random_data(she[i].pass_salt, HIVE_SALT_LEN);
                get_random_data(she[i].key, HIVE_KEY_LEN);
                get_random_data(she[i].key_salt, HIVE_SALT_LEN);
                get_random_data(she[i].key_digest, HIVE_SALT_LEN);
            }
        }
    } else {
        /* Decrypt the actual encryption keys. */
        for (i = 0; i < hive_worlds; ++i) {
            decrypt(she[i].key, hive_keys[i], HIVE_KEY_LEN, hive_pass_keys[i], 0);
        }
    }

    /* Clean the pass key memory. */
    for (i = 0; i < HIVE_MAX_WORLDS; ++i) {
        memset(hive_pass_keys[i], 0, HIVE_KEY_LEN);
    }
}

/* Decrypt and verify the HiVE header magic text with the given keys. */
int verify_keys(struct hive_header_entry *she) {

    char digest[HIVE_DIGEST_LEN];
    unsigned i;
    unsigned r;

    r = 1;
    for (i = 0; i < hive_worlds; ++i) {
        digest_key(hive_keys[i], HIVE_KEY_LEN, digest, HIVE_DIGEST_LEN, she[i].key_salt, HIVE_SALT_LEN);
        if(memcmp(digest, she[i].key_digest, HIVE_DIGEST_LEN) != 0) {
            r = 0;
        }
    }

    return r;
}

/* Clean the key memory. */
void cleanup_keys() {
    unsigned i;

    for (i = 0; i < HIVE_MAX_WORLDS; ++i) {
        memset(hive_keys[i], 0, HIVE_KEY_LEN);
    }
}

/* Constructs names and paths for the mapped devices. */
void get_names(char *public_dev_name, char *hidden_dev_name,
               char **public_dev_path,
               char **hidden_dev_path,
               char **map_hive_name,
               char **public_dev_map_name,
               char **hidden_dev_map_name,
               char **public_dev_map_path,
               char **hidden_dev_map_path)
{
    *public_dev_path = malloc(strlen(HIVE_DEV_PATH) + strlen(public_dev_name) + 1);
    *hidden_dev_path = malloc(strlen(HIVE_DEV_PATH) + strlen(public_dev_name) + 1);
    strcpy(*public_dev_path, HIVE_DEV_PATH);
    strcat(*public_dev_path, public_dev_name);
    strcpy(*hidden_dev_path, HIVE_DEV_PATH);
    strcat(*hidden_dev_path, hidden_dev_name);

    /*
     * The user isn't supposed to mess with the map devices, so use a random
     * name to avoid a clash with user-specified HiVE names.
     */
    *map_hive_name = get_random_name();

    *public_dev_map_name = malloc(strlen(HIVE_MAP_DEV_PREFIX) + strlen(*map_hive_name) +
                                  strlen(HIVE_PUBLIC_DEV_SUFFIX) + 1);
    strcpy(*public_dev_map_name, HIVE_MAP_DEV_PREFIX);
    strcat(*public_dev_map_name, *map_hive_name);
    strcat(*public_dev_map_name, HIVE_PUBLIC_DEV_SUFFIX);

    *hidden_dev_map_name = malloc(strlen(HIVE_MAP_DEV_PREFIX) + strlen(*map_hive_name) +
                                  strlen(HIVE_HIDDEN_DEV_SUFFIX) + 1);
    strcpy(*hidden_dev_map_name, HIVE_MAP_DEV_PREFIX);
    strcat(*hidden_dev_map_name, *map_hive_name);
    strcat(*hidden_dev_map_name, HIVE_HIDDEN_DEV_SUFFIX);

    *public_dev_map_path = malloc(strlen(HIVE_DEV_PATH) + strlen(*public_dev_map_name) + 1);
    strcpy(*public_dev_map_path, HIVE_DEV_PATH);
    strcat(*public_dev_map_path, *public_dev_map_name);

    *hidden_dev_map_path = malloc(strlen(HIVE_DEV_PATH) + strlen(*hidden_dev_map_name) + 1);
    strcpy(*hidden_dev_map_path, HIVE_DEV_PATH);
    strcat(*hidden_dev_map_path, *hidden_dev_map_name);
}


/*
 * Meta block functions.
 */

/* Encrypts an arbitrarily large buffer and writes it to the device. */
void encrypt_and_write_sectors(int fd, char *src, unsigned count, char *key, char *iv) {

    unsigned c;
    int len;
    char *dst;
    EVP_CIPHER_CTX *ctx;

    if (count == 0) {
        return;
    }

    dst = malloc(HIVE_IO_SIZE * HIVE_SECTOR_LEN);

    ctx = start_encrypt(key, iv);
    while (count > 0) {
        c = (count >= HIVE_IO_SIZE) ? HIVE_IO_SIZE : count;
        len = do_encrypt(ctx, src, dst, c * HIVE_SECTOR_LEN);
        write_bytes(fd, dst, len);
        count -= c;
    }
    len = finish_encrypt(ctx, dst);
    write_bytes(fd, dst, len);

    free(dst);
}

/* Writes map entries to a different HiVE for a recursive HiVE configuration. */
void write_map_to_other_hive(char *dev, u64 map_index, u64 map_len) {

    int fd;
    u64 *map;
    unsigned c;
    unsigned i;

    map = malloc(HIVE_IO_SIZE * HIVE_SECTOR_LEN);

    if ((fd = open(dev, O_RDWR)) == -1) {
        die("Cannot open device %s\n", dev);
    }

    while (map_len > 0) {
        c = (map_len >= HIVE_IO_SIZE) ? HIVE_IO_SIZE : map_len;

        for (i = 0; i < c * HIVE_SECTOR_LEN / sizeof(u64); ++i) {
            map[i] = map_index++;
        }
        write_sectors(fd, (char *) map, c);
        map_len -= c;
    }

    close(fd);
    free(map);
}

/* Writes encrypted map entries for a HiVE that keeps the forward maps in memory. */
void write_map_to_this_hive(int fd, u64 map_index, u64 map_len, char *key, char *iv) {

    u64 *map;
    char *buf;
    unsigned c;
    int len;
    unsigned i;

    EVP_CIPHER_CTX *ctx;

    map = malloc(HIVE_IO_SIZE * HIVE_SECTOR_LEN);
    buf = malloc(HIVE_IO_SIZE * HIVE_SECTOR_LEN);

    ctx = start_encrypt(key, iv);
    while (map_len > 0) {
        c = (map_len >= HIVE_IO_SIZE) ? HIVE_IO_SIZE : map_len;

        for (i = 0; i < c * HIVE_SECTOR_LEN / sizeof(u64); ++i) {
            map[i] = map_index++;
        }
        len = do_encrypt(ctx, (char *) map, buf, c * HIVE_SECTOR_LEN);
        write_bytes(fd, buf, len);
        map_len -= c;
    }
    len = finish_encrypt(ctx, buf);
    write_bytes(fd, buf, len);

    free(map);
    free(buf);
}


/* Encrypts a meta block. */
void encrypt_meta(struct hive_meta_block *m, char *key) {

    char *src;
    char dst[16];
    u64 len;

    len = sizeof(*m) - HIVE_IV_LEN;
    src = ((char *) m) + HIVE_IV_LEN;
    encrypt(src, dst, len, key, m->iv);

    memcpy(src, dst, len);
}

/* Writes meta sectors for used data blocks. */
void write_meta(int fd, u64 rmap_index, u64 world, u64 len, char *key) {

    struct hive_meta_block *meta;
    unsigned meta_per_sector;
    u64 i;
    u64 j;

    meta_per_sector = HIVE_SECTOR_LEN / sizeof(*meta);
    meta = malloc(meta_per_sector * sizeof(*meta));

    for (i = 0; i < len;) {
        for (j = 0; j < meta_per_sector && i < len; ++j, ++i) {
            get_random_data(meta[j].iv, HIVE_IV_LEN);
            meta[j].sector = rmap_index++;
            meta[j].world = world;
            encrypt_meta(meta + j, key);
        }
        write_sectors(fd, (char *) meta, 1);
    }

    free(meta);
}

/* Writes randomized meta blocks for unused data sectors. */
void write_random_meta(int fd, u64 len) {

    struct hive_meta_block *meta;
    unsigned meta_per_sector;
    u64 i;
    u64 j;

    meta_per_sector = HIVE_SECTOR_LEN / sizeof(*meta);
    meta = malloc(meta_per_sector * sizeof(*meta));

    for (i = 0; i < len;) {
        for (j = 0; j < meta_per_sector && i < len; ++j, ++i) {
            get_random_data(meta[j].iv, HIVE_IV_LEN);
            get_random_data((char *) &meta[j].sector, sizeof(meta[j].sector));
            get_random_data((char *) &meta[j].world, sizeof(meta[j].world));
        }
        write_sectors(fd, (char *) meta, 1);
    }

    free(meta);
}

/*
 * Actual commands.
 */

/* Closes a single HiVE world by device-mapper name. */
int close_hive(char *name) {

    struct dm_task *dmt;
    u32 cookie = 0;
    u16 udev_flags = 0;
    int is_success = 0;

#ifdef HIVE_NO_UDEV
    udev_flags |= DM_UDEV_DISABLE_DM_RULES_FLAG | DM_UDEV_DISABLE_SUBSYSTEM_RULES_FLAG;
    dm_udev_set_sync_support(0);
#endif

    if (!(dmt = dm_task_create(DM_DEVICE_REMOVE))) {
        print_red("DEBUG: Cannot create dm_task\n");
        return 0;
    }

    if (!dm_task_set_name(dmt, name)) {
        print_red("DEBUG: Cannot set device name\n");
        goto out;
    }

    if (!dm_task_set_cookie(dmt, &cookie, udev_flags)) {
        goto out;
    }

    dm_task_retry_remove(dmt);

    if (!dm_task_run(dmt)) {
        print_red("DEBUG: Cannot issue ioctl\n");
        goto out;
    }

    is_success = 1;

#ifdef HIVE_NO_UDEV
    dm_mknodes(name);
#endif

out:
    dm_udev_wait(cookie);
    dm_task_destroy(dmt);

    return is_success;
}

void _do_close(char **names, int len) {

    int i;
    char *name;

    for (i = len - 1; i >= 0; --i) {
        name = rindex(names[i], '/') + 1;
        if (!close_hive(name)) {
            print_red("DEBUG: Cannot close %s\n", name);
        }
    }
}

/* Removes all worlds in a given HiVE instance. */
void do_close(char *hive_name) {

    char *buf;
    char *tok_buf;
    char *tok;
    unsigned len;

    char *dev_names[HIVE_MAX_WORLDS];
    char *map_names[HIVE_MAX_WORLDS];

    int dev_count;
    int map_count;

    unsigned i;
    int done;

    sync();

    /* Read the HiVE proc file. */
    buf = read_text_file(HIVE_PROC_FILE, &len);
    if (!len) {
        print_red("There are no HiVE devices open.\n");
        return;
    }

    buf[len - 1] = '\0';
    tok_buf = buf;

    /* Check all open HiVE instances. */
    done = 0;
    while (!done && (tok = strsep(&tok_buf, " "))) {

        if (strcmp(tok, hive_name) != 0) {
            /* This is not what we are looking for. */
            strsep(&tok_buf, "\n"); /* Skip to the end of line. */
        }
        else {
            /*
             * Found! The kernel module does not know anything about the mapping
             * order of HiVE devices, so the below code must remove all
             * mappings in the correct order:
             *
             * Hidden data device > Public data device > Hidden map device > Public map device
             */
            dev_count = atoi(strsep(&tok_buf, " "));
            for (i = 0; i < dev_count; ++i) {
                dev_names[i] = strsep(&tok_buf, " ");
            }
            _do_close(dev_names, dev_count);

            map_count = atoi(strsep(&tok_buf, " "));
            for (i = 0; i < map_count; ++i) {
                map_names[i] = strsep(&tok_buf, " ");
            }
            _do_close(map_names, map_count);
            done = 1;
        }
    }

    /* Not found. */
    if (!done) {
        print_red("No HiVE named \"%s\"\n", hive_name);
    }

    free(buf);
}

/* Opens a single HiVE world. */
int open_hive(char *dev_path, char *hive_name, char *dm_name, char *dm_path, u64 len, u64 start, int is_single_world, char *map_dev) {

    struct dm_task *dmt;
    u32 cookie = 0;
    u16 udev_flags = 0;
    int is_success = 0;
    char param[4096];
    char *keys[HIVE_MAX_WORLDS];
    unsigned i;

#ifdef HIVE_NO_UDEV
    udev_flags |= DM_UDEV_DISABLE_DM_RULES_FLAG | DM_UDEV_DISABLE_SUBSYSTEM_RULES_FLAG;
    dm_udev_set_sync_support(0);
#endif

    /* Hex encode the key bytes. */
    for(i = 0; i < HIVE_MAX_WORLDS; ++i) {
        keys[i] = hex_encode(hive_keys[i], HIVE_KEY_LEN);
    }

    /* These parameters will be passed to the kernel module. */
    snprintf(param, 4096, "%s %s %s %llu %d %s %s %s\n",
             dev_path, hive_name, dm_path, start,
             is_single_world, keys[WORLD_0], keys[WORLD_1], map_dev);

    if (!(dmt = dm_task_create(DM_DEVICE_CREATE))) {
        print_red("DEBUG: Cannot create dm_task\n");
        return 0;
    }

    if (!dm_task_set_name(dmt, dm_name)) {
        print_red("DEBUG: Cannot set device name\n");
        goto out;
    }

    if (!dm_task_add_target(dmt, 0, len * (HIVE_SECTOR_LEN / 512), HIVE_TARGET, param)) {
        goto out;
    }

    if (!dm_task_set_add_node(dmt, DM_ADD_NODE_ON_CREATE)) {
        print_red("DEBUG: Cannot add node\n");
        goto out;
    }

    if (!dm_task_set_cookie(dmt, &cookie, udev_flags)) {
        print_red("DEBUG: Cannot set cookie\n");
        goto out;
    }

    if (!dm_task_run(dmt)) {
        print_red("DEBUG: Cannot issue ioctl\n");
        goto out;
    }

#ifdef HIVE_NO_UDEV
    dm_mknodes(dm_name);
#endif

is_success = 1;

out:
    dm_udev_wait(cookie);
    dm_task_destroy(dmt);

    /* Clean the keys. */
    memset(param, 0, 4096);
    for(i = 0; i < HIVE_MAX_WORLDS; ++i) {
        memset(keys[i], 0, HIVE_KEY_LEN * 2);
        free(keys[i]);
    }

    return is_success;
}

/* Open shortcuts for map and data worlds. */
int open_map_hive(char *dev_path, char *hive_name, char *dm_name, char *dm_path,  u64 len, u64 start, int is_single_world) {
    return open_hive(dev_path, hive_name, dm_name, dm_path, len, start, is_single_world, HIVE_NO_MAP_HIVE);
}

int open_data_hive(char *dev_path, char *hive_name, char *dm_name, char *dm_path, u64 len, u64 start, int is_single_world, char *map_hive) {
    return open_hive(dev_path, hive_name, dm_name, dm_path, len, start, is_single_world, map_hive);
}

/* Open a complete HiVE instance including up to 2 data worlds and 2 map worlds. */
void do_open(char *dev_path, char *hive_name, char *public_name, char *hidden_name, int is_single_world) {

    int fd;
    struct hive_header_entry *she_data;
    struct hive_header_entry *she_map;
    char *hive_headers_data;
    char *hive_headers_map;
    char *public_dev_path;
    char *hidden_dev_path;
    char *public_dev_map_name;
    char *hidden_dev_map_name;
    char *public_dev_map_path;
    char *hidden_dev_map_path;
    char *map_hive_name;

    /* Warning! You will lose all hidden world data! */
    if (is_single_world) {
        char buf[4];
        print_red("You are opening this device in single-volume mode.\n"
                  "This will destroy your hidden volume, if there is one.\n"
                  "Do you want to proceed? (Type \"YES\") ");
        if(scanf("%3s", buf) != 1 || strcmp(buf, "YES") != 0) {
            die("Aborting\n");
        }
        print_green("Proceeding...\n");
    }

    /* Open device. */
    if ((fd = open(dev_path, O_RDWR)) == -1) {
        die("Cannot open device %s\n", dev_path);
    }

    /* Read headers. */
    hive_headers_data = malloc(HIVE_HEADER_LEN * HIVE_SECTOR_LEN * 2);
    hive_headers_map = hive_headers_data + (HIVE_HEADER_LEN * HIVE_SECTOR_LEN);

    read_sectors(fd, hive_headers_data, HIVE_HEADER_LEN);
    she_data = (struct hive_header_entry *) hive_headers_data;

    /* Get passwords from user and check if they are correct. */
    get_keys(HIVE_OPEN, she_data);
    if(!verify_keys(she_data)) {
        print_red("Incorrect password(s)!\n");
        goto free_headers;
    }

    lseek(fd, she_data[0].next_hive * HIVE_SECTOR_LEN, SEEK_SET);
    read_sectors(fd, hive_headers_map, HIVE_HEADER_LEN);
    she_map = (struct hive_header_entry *) hive_headers_map;

    get_names(public_name, hidden_name,
              &public_dev_path, &hidden_dev_path,
              &map_hive_name,
              &public_dev_map_name, &hidden_dev_map_name,
              &public_dev_map_path, &hidden_dev_map_path);

    /* Start opening. Order is important. */
    if (!open_map_hive(dev_path, map_hive_name, public_dev_map_name, public_dev_map_path,
                         she_map[WORLD_0].world_len, she_data[WORLD_0].next_hive, is_single_world)) {
        print_red("Cannot open public map device!\n");
        goto free_names;
    }

    if (!is_single_world && !open_map_hive(dev_path, map_hive_name, hidden_dev_map_name, hidden_dev_map_path,
                         she_map[WORLD_1].world_len, she_data[WORLD_0].next_hive, is_single_world)) {
        print_red("Cannot open hidden map device!\n");
        goto close_public_map;
    }

    if (!open_data_hive(dev_path, hive_name, public_name, public_dev_path,
                          she_data[WORLD_0].world_len, 0, is_single_world, map_hive_name)) {
        print_red("Cannot open public device!");
        goto close_hidden_map;
    }

    if (!is_single_world && !open_data_hive(dev_path, hive_name, hidden_name, hidden_dev_path,
                          she_data[WORLD_1].world_len, 0, is_single_world, map_hive_name)) {
        print_red("Cannot open hidden device!");
        goto close_public_data;
    }

    print_green("Success!\n");
    goto success;

close_public_data:
    close_hive(public_name);
close_hidden_map:
    close_hive(hidden_dev_map_name);
close_public_map:
    close_hive(public_dev_map_name);
success:
free_names:
    free(map_hive_name);
    free(public_dev_path);
    free(hidden_dev_path);
    free(public_dev_map_name);
    free(hidden_dev_map_name);
    free(public_dev_map_path);
    free(hidden_dev_map_path);
free_headers:
    cleanup_keys();
    free(hive_headers_data);
    close(fd);
}

/* Creates a single HiVE with up to two worlds. */
void create_hive(int fd, u64 start, u64 sectors,
                 char *public_map_dev, char *hidden_map_dev, char *hive_headers) {
    u64 data_len;
    u64 meta_len;
    u64 map_len;
    u64 header_map_len;
    struct hive_header_entry *she;
    char *stash;
    u64 i;

    lseek(fd, start * HIVE_SECTOR_LEN, SEEK_SET);

    data_len = get_data_sectors(sectors);
    meta_len = get_meta_sectors(sectors);
    map_len = get_map_sectors(sectors);

    if (!public_map_dev || !hidden_map_dev) {
        header_map_len = map_len;
    }
    else {
        header_map_len = 0;
    }

#ifdef HIVE_DEBUG
    print_green("Start sector: %llu\n", start);
    print_green("Total data sectors: %llu\n", data_len);
    print_green("-> Data sectors per volume: %llu\n", sectors);
    print_green("Translation map sectors per volume: %llu\n", map_len);
    print_green("Header map sectors: %llu\n", header_map_len);
    print_green("Meta sectors: %llu\n", meta_len);
#endif

    /*
     * Set up HiVE headers. Assuming 2 worlds for now.
     */

    /* World 0. */
    she = (struct hive_header_entry *) hive_headers;
    she[WORLD_0].stash_start = HIVE_HEADER_LEN;
    she[WORLD_0].stash_len = HIVE_STASH_LEN;
    she[WORLD_0].stash_map_start = she[WORLD_0].stash_start + she[WORLD_0].stash_len + HIVE_STASH_LEN;
    she[WORLD_0].stash_map_len = HIVE_STASH_MAP_LEN;
    she[WORLD_0].trans_start = she[WORLD_0].stash_map_start + she[WORLD_0].stash_map_len + HIVE_STASH_MAP_LEN;
    she[WORLD_0].trans_len = header_map_len;
    she[WORLD_0].meta_start = she[WORLD_0].trans_start + (2 * header_map_len);
    she[WORLD_0].data_start = she[WORLD_0].meta_start + meta_len;
    she[WORLD_0].data_len = data_len;
    she[WORLD_0].world_len = sectors;
    she[WORLD_0].next_hive = she[WORLD_0].data_start + data_len;
    get_random_data(she[WORLD_0].iv, HIVE_IV_LEN);

    /* World 1. */
    she[WORLD_1].stash_start = she[WORLD_0].stash_start + HIVE_STASH_LEN;
    she[WORLD_1].stash_len = HIVE_STASH_LEN;
    she[WORLD_1].stash_map_start = she[WORLD_0].stash_map_start + HIVE_STASH_MAP_LEN;
    she[WORLD_1].stash_map_len = HIVE_STASH_MAP_LEN;
    she[WORLD_1].trans_start = she[WORLD_0].trans_start + header_map_len;
    she[WORLD_1].trans_len = header_map_len;
    she[WORLD_1].meta_start = she[WORLD_0].trans_start + (2 * header_map_len);
    she[WORLD_1].data_start = she[WORLD_0].meta_start + meta_len;
    she[WORLD_1].data_len = data_len;
    she[WORLD_1].world_len = sectors;
    she[WORLD_1].next_hive = 0;
    get_random_data(she[WORLD_1].iv, HIVE_IV_LEN);

    /* Write them. */
    write_sectors(fd, hive_headers, HIVE_HEADER_LEN);

    /*
     * Set up stash.
     */
    stash = malloc(HIVE_SECTOR_LEN * HIVE_STASH_MAP_LEN);
    for (i = 0; i < 2 * HIVE_STASH_LEN; ++i) {
        get_random_data(stash, HIVE_SECTOR_LEN);
        write_sectors(fd, stash, 1);
    }

    memset(stash, 0xff, HIVE_SECTOR_LEN * HIVE_STASH_MAP_LEN);
    encrypt_and_write_sectors(fd, stash, HIVE_STASH_MAP_LEN, hive_keys[WORLD_0], she[WORLD_0].iv);
    encrypt_and_write_sectors(fd, stash, HIVE_STASH_MAP_LEN, hive_keys[WORLD_1], she[WORLD_1].iv);

    free(stash);

    /*
     * Set translation maps.
     */

    if (!public_map_dev || !hidden_map_dev) {
        /* This is a map HiVE. Translation maps will be kept in memory. */
        write_map_to_this_hive(fd, 0, map_len, hive_keys[WORLD_0], she[WORLD_0].iv);
        write_map_to_this_hive(fd, sectors, map_len, hive_keys[WORLD_1], she[WORLD_1].iv);
    } else {
        /*
         * This is a data HiVE, there are separate map worlds. Write to
         * translation map to map worlds.
         */
        write_map_to_other_hive(public_map_dev, 0, map_len);
        write_map_to_other_hive(hidden_map_dev, sectors, map_len);
    }

    /*
     * Set up metadata.
     */

    write_meta(fd, 0, WORLD_0, data_len / 4, hive_keys[WORLD_0]);
    write_meta(fd, sectors, WORLD_1, data_len / 4, hive_keys[WORLD_1]);
    write_random_meta(fd, data_len / 2);

    sync();

    /* All done. */
    print_green("Done!\n\n");
}

/* Create a complete HiVE instance, with a data HiVE and a map HiVE, up to two worlds each. */
void do_create(char *dev_path, char *public_name, char *hidden_name) {

    int fd;
    u64 dev_size;
    u64 sectors_data_hive;
    u64 sectors_map_hive;
    unsigned size_correction;
    char *hive_headers;
    char *public_dev_path;
    char *hidden_dev_path;
    char *public_dev_map_name;
    char *hidden_dev_map_name;
    char *public_dev_map_path;
    char *hidden_dev_map_path;
    char *map_hive_name;

    /* Open device. */
    if ((fd = open(dev_path, O_RDWR)) == -1) {
        die("Cannot open device %s\n", dev_path);
    }

    init_random();
    init_crypto();

    /* ioctl for device size. */
    if (ioctl(fd, BLKGETSIZE64, &dev_size) == -1) {
        die("Cannot determine size of device %s\n", dev_path);
    }

#ifdef HIVE_DEBUG
    print_green("Device size is %llu bytes\n", dev_size);
    print_green("-> Sectors: %llu \n\n", dev_size / HIVE_SECTOR_LEN);
#endif

    /* Compute the HiVE device size. */
    sectors_data_hive = (HIVE_SECTOR_LEN * (dev_size -
                                              (2 * HIVE_HEADER_LEN * HIVE_SECTOR_LEN) -
                                              (4 * HIVE_STASH_LEN * HIVE_SECTOR_LEN) -
                                              (4 * HIVE_STASH_MAP_LEN * HIVE_SECTOR_LEN)))
		        / (HIVE_SECTOR_LEN * (4 * HIVE_SECTOR_LEN + 160) + 1152);

    sectors_map_hive = div_ceil(8 * sectors_data_hive, HIVE_SECTOR_LEN);

    /*
     * Did the wasted portions of metadata sectors affect the data storage? If
     * so, correct the world sizes.
     */
    size_correction = get_size_correction(dev_size / HIVE_SECTOR_LEN, sectors_data_hive, sectors_map_hive);
    if (size_correction) {
        sectors_data_hive -= size_correction;
        sectors_map_hive = div_ceil(8 * sectors_data_hive, HIVE_SECTOR_LEN);
    }

#ifdef HIVE_DEBUG
    print_green("Sectors per data volume: %llu\n", sectors_data_hive);
    print_green("Sectors per map volume: %llu\n", sectors_map_hive);
    print_green("-> Size correction is: %u\n\n", size_correction);
#endif

    /* Create empty headers. */
    hive_headers = malloc(HIVE_HEADER_LEN * HIVE_SECTOR_LEN);
    memset(hive_headers, 0, HIVE_HEADER_LEN * HIVE_SECTOR_LEN);

    /* Prompt user for passwords. */
    get_keys(HIVE_CREATE, (struct hive_header_entry *) hive_headers);

    /* Create the map HiVE. */
    print_green("Creating map volumes...\n");
    create_hive(fd, get_data_hive_len(sectors_data_hive), sectors_map_hive, NULL, NULL, hive_headers);
    close(fd); /* Close the device so that everything is actually written. */

    get_names(public_name, hidden_name,
              &public_dev_path, &hidden_dev_path,
              &map_hive_name,
              &public_dev_map_name, &hidden_dev_map_name,
              &public_dev_map_path, &hidden_dev_map_path);

    /* Temporarily open the map HiVE, data HiVE's translation maps will be written there. */
    if (!open_map_hive(dev_path, map_hive_name, public_dev_map_name, public_dev_map_path, sectors_map_hive,
                         get_data_hive_len(sectors_data_hive), 0)) {
        print_red("Cannot open public map device!\n");
    }

    if (!open_map_hive(dev_path, map_hive_name, hidden_dev_map_name, hidden_dev_map_path, sectors_map_hive,
                         get_data_hive_len(sectors_data_hive), 0)) {
        print_red("Cannot open hidden map device!\n");
    }

    if ((fd = open(dev_path, O_RDWR)) == -1) {
        cleanup_keys();
        die("Cannot open device %s\n", dev_path);
    }

    /* Create the data HiVE. */
    print_green("Creating data volumes...\n");
    create_hive(fd, 0, sectors_data_hive, public_dev_map_path, hidden_dev_map_path, hive_headers);

    /* We can now close the map HiVES. */
    if (!close_hive(hidden_dev_map_name)) {
        print_red("DEBUG: Cannot close %s\n", hidden_dev_map_name);
    }

    if (!close_hive(public_dev_map_name)) {
        print_red("DEBUG: Cannot close %s\n", public_dev_map_name);
    }

    close(fd);

    free(hive_headers);
    free(map_hive_name);
    free(public_dev_path);
    free(hidden_dev_path);
    free(public_dev_map_name);
    free(hidden_dev_map_name);
    free(public_dev_map_path);
    free(hidden_dev_map_path);

    cleanup_keys();
    cleanup_crypto();
    cleanup_random();
}

/* Reads the HiVE proc file and lists all open HiVE instances. */
void do_list() {

    char *buf;
    char *tok_buf;
    char *tok;
    unsigned len;

    int dev_count;
    int map_count;

    unsigned i;

    buf = read_text_file(HIVE_PROC_FILE, &len);
    if (!len)
        return;

    buf[len - 1] = '\0';
    tok_buf = buf;

    while ((tok = strsep(&tok_buf, " "))) {
        print_green("\n%s\n", tok);
        print_green("  Devices: \n");
        dev_count = atoi(strsep(&tok_buf, " "));
        for (i = 0; i < dev_count; ++i) {
            printf("   --> Volume %u: %s\n", i, strsep(&tok_buf, " "));
        }

        map_count = atoi(strsep(&tok_buf, " "));
        if (map_count) {
            print_green("  Maps (Don't touch these!): \n");
        }
        for (i = 0; i < map_count; ++i) {
            printf("   --> Map %u: %s\n", i, strsep(&tok_buf, " "));
        }
        strsep(&tok_buf, "\n"); /* Skip to the end of line. */
    }

    free(buf);
}
