/*
 * Copyright (C) 2014 Erik-Oliver Blass
 *                    Travis Mayberry
 *                    Guevara Noubir
 *                    Kaan Onarlioglu <onarliog@ccs.neu.edu>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * main.c, ver.2014.11.03
 *
 * HiVE Userland Tool.
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <argp.h>

#include "hive.h"

/*
 * Argument parsing stuff, using argp.
 */

/* Command arguments. */
#define COMMAND_CREATE "create"
#define COMMAND_OPEN "open"
#define COMMAND_CLOSE "close"
#define COMMAND_LIST "list"

const char *argp_program_version = "HiVE ver.2014.07.23";
const char *argp_program_bug_address = "<onarliog@ccs.neu.edu>";
static const char doc[] = "Create, open, close, list HiVE devices.";
static const char args_doc[] =
    COMMAND_CREATE  " <block-device>\n"
    COMMAND_OPEN    " <block-device> <hive-name>\n"
    COMMAND_CLOSE   " <hive-name>\n"
    COMMAND_LIST    "\n";

static struct argp_option options[] = {
    {"public-volume", 'p', "<public-volume-name>", 0, "Public volume name"},
    {"hidden-volume", 'h', "<hidden-volume-name>", 0, "Hidden volume name"},
    {"single-volume", 's', 0, 0, "Only create/open a public volume"},
    {0}
};

struct arguments {
    char *args[3];
    char *hive_dev;
    char *public_dev;
    char *hidden_dev;
    int is_single_world;
};

static error_t parse_arguments(int key, char *arg, struct argp_state *state) {

    struct arguments *arguments = state->input;

    switch (key) {
    case 'p':
        arguments->public_dev = arg;
        break;
    case 'h':
        arguments->hidden_dev = arg;
        break;
    case 's':
        arguments->is_single_world = 1;
        break;
    case ARGP_KEY_ARG:
        if ((state->arg_num > 2 && strcmp(arguments->args[0], COMMAND_OPEN) == 0) ||
            (state->arg_num > 1 && strcmp(arguments->args[0], COMMAND_CREATE) == 0) ||
            (state->arg_num > 1 && strcmp(arguments->args[0], COMMAND_CLOSE) == 0) ||
            (state->arg_num > 0 && strcmp(arguments->args[0], COMMAND_LIST) == 0)){
            /* Too many arguments. */
            argp_usage(state);
        }
        arguments->args[state->arg_num] = arg;
        break;
    case ARGP_KEY_END:
        if (state->arg_num == 0 ||
            (state->arg_num < 1 && strcmp(arguments->args[0], COMMAND_LIST) == 0) ||
            (state->arg_num < 2 && strcmp(arguments->args[0], COMMAND_CREATE) == 0) ||
            (state->arg_num < 2 && strcmp(arguments->args[0], COMMAND_CLOSE) == 0) ||
            (state->arg_num < 3 && strcmp(arguments->args[0], COMMAND_OPEN) == 0)) {
            /* Missing arguments. */
            argp_usage(state);
        }
        break;
    default:
        return ARGP_ERR_UNKNOWN;
    }
    return 0;
}

static struct argp arg_parser = {options, parse_arguments, args_doc, doc};
struct arguments arguments;

int main(int argc, char **argv) {

    assert((sizeof(struct hive_header_entry) % HIVE_IV_LEN) == 0);
    assert((HIVE_SECTOR_LEN % sizeof(struct hive_meta_block)) == 0);

    /* Default arguments. */
    arguments.public_dev = "hive-public";
    arguments.hidden_dev = "hive-hidden";
    arguments.is_single_world = 0;

    /* Parse arguments. */
    argp_parse(&arg_parser, argc, argv, 0, 0, &arguments);

    if (arguments.is_single_world) {
        hive_worlds = 1;
    } else {
        hive_worlds = HIVE_MAX_WORLDS;
    }

    /* Command dispatcher. */
    if (strcmp(arguments.args[0], COMMAND_CREATE) == 0) {

        print_green("Creating HiVE on %s\n", arguments.args[1]);
        do_create(arguments.args[1], arguments.public_dev, arguments.hidden_dev);
    }
    else if (strcmp(arguments.args[0], COMMAND_OPEN) == 0) {

        print_green("Opening HiVE device %s on %s\n", arguments.args[2], arguments.args[1]);
        print_green("Public volume is /dev/mapper/%s\n", arguments.public_dev);
        if (!arguments.is_single_world) {
            print_green("Hidden volume is /dev/mapper/%s\n", arguments.hidden_dev);
        }
        do_open(arguments.args[1], arguments.args[2], arguments.public_dev, arguments.hidden_dev, arguments.is_single_world);
    }
    else if (strcmp(arguments.args[0], COMMAND_CLOSE) == 0) {

        print_green("Closing HiVE device %s \n", arguments.args[1]);
        do_close(arguments.args[1]);
    }
    else if (strcmp(arguments.args[0], COMMAND_LIST) == 0) {
        print_green("Listing open HIVE devices.\n");
        do_list();
    }
    else {
        die("Unknown command: %s\n", arguments.args[0]);
    }

    return 0;
}
